%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Script for running the optimization for finding coefficients of Bzier 
%%% polynom and other desired parameters.
%%% Script for solving the 2nd-order differential equation of the reduced
%%% dynamics and plotting joint angles, velocities and torques with respect
%%% to time.
%%% Script for comparing measured data from experiments to simulated data.
%%% Script for making data input file for KUKA lwr with 2ms sample-time.
%%% More comments in each section.
%%%
%%% Brede Lvik Lillehammer          NTNU        03.06.14
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

clear all
set(0,'DefaultAxesXGrid','on','DefaultAxesYGrid','on')
global I1x I1y I1z I2x I2y I2z I3x I3y I3z 
global lmx1 lmy1 lmz1 lmx2 lmy2 lmz2 lmx3 lmy3 lmz3 
global l0 l1 l2 l3
global m1 m2 m3 kspr g
global psi
global theta_b theta_e
global samples lb ub
global coeff_degree
global VHC_coeff

m_6 = 1.5724; m_gripper = 2.860731; m_object = 0.08;
m1 = 4.60714; m2 = 3.89944;    m3 = m_6+m_gripper+m_object;

% Calculating CoM for link 3
lmx3 = 1/m3*(m_gripper*(-0.018567));
lmy3 = 1/m3*(m_6*0.00408+m_gripper*0.01526);
lmz3 = 1/m3*(m_6*0.00157+m_gripper*(0.078+0.083407)+m_object*(0.078+0.092+0.035));

lmx1 = 0;        lmx2 = 0;
lmy1 = -0.02522; lmy2 = 0.0222;
lmz1 = 0.198;    lmz2 = 0.1561;

l1 = 0.4;        l2 = 0.39728;   l3 = 0.240;     l0 = 0.3105; 

I1x = 0.0086628; I2x = 0.007843;
I1y = 0.096343;  I2y = 0.063329;
I1z = 0.2743713; I2z = 0.15196;

% Calculating inertia for link three, consisting of link three of robot,
% gripper and object to be thrown. Weight of object can vary.
I3x = 0.0102278+m3*lmx3^2;
I3y = 0.0103661+m3*lmy3^2;
I3z = 0.0123651+m3*lmz3^2;

kspr = 20; g = 9.81;
psi = pi/4;

%% Setting upper and lower bounds for parameters used in optimization search.
% z0 is vector for control points of Bzier for VHC 1 and 2 are z0=[a0;b0]

coeff_degree = length(z0)/2-1;
lb = -3*ones(length(z0),1);
ub =  5*ones(length(z0),1);
lb(coeff_degree*2+3:coeff_degree*2+6) = [-1; -1; 0;  0];
ub(coeff_degree*2+3:coeff_degree*2+6) = [1; 1; 100; pi];

%% Just for checking start and release point of theta, usefull for finding new trajectories
samples = 100;

wrist_joint_start = 1;
wrist_joint_end = 0;

theta_b = l1*cos(z0(1)-psi)+l2*cos(z0(1)+z0(coeff_degree+2)-psi)+l3*cos(z0(1)+z0(coeff_degree+2)+wrist_joint_start-psi)
theta_e = l1*cos(z0(coeff_degree+1)-psi)+l2*cos(z0(coeff_degree+1)+z0(end)-psi)+l3*cos(z0(coeff_degree+1)+z0(end)+wrist_joint_end-psi)

theta_interval = linspace(theta_b,theta_e,samples);
%% Optimization
warning('off','all');
options = optimoptions(@fmincon,'Algorithm','sqp','Display','iter-detailed','Diagnostics','off',...
    'MaxFunEvals',400,'TolFun',1e-6,'TolX',1e-9,'TolCon',5e-9);
% q0 = [z0;theta_b;theta_e;kspr;psi];
% q0 = coeff_vector;
% theta_dot_sqrd_vec = theta_dot_sqrd(z0);

% The optimization
tic 
[coeff_vector,fval,exitflag,output,lambda,grad,hessian] = optim(q0,options);
toc

% For testing if the starting point is feasible, or making a starting point
% feasible. Open nonLinCon.m for more info.
coeff_vector = fmincon(@(q)0,q0,[],[],[],[],lb,ub,@nonLinCon,options);

% Currently used scalar parameters used for the optimization search
z0 = coeff_vector(1:end-4); 
theta_b = coeff_vector(end-3);
theta_e = coeff_vector(end-2);
kspr = coeff_vector(end-1); 
psi = coeff_vector(end);

%% Some different kind of plots

% Plotting alpha3
figure
plot(theta_interval,al3(theta_interval,z0))
title('alpha')

theta_dot_sqrd_vec = theta_dot_sqrd(z0)';
D2theta = 1/2*(diff(theta_dot_sqrd_vec)./diff(theta_interval));
red_eq = al3(theta_interval(2:end),z0).*D2theta+be3(theta_interval(2:end),z0).*theta_dot_sqrd_vec(2:end)+gam3(theta_interval(2:end),z0);

% Plotting theta_dot with respect to the interval theta
figure
plot(theta_interval,sqrt(theta_dot_sqrd_vec),'r')
title('theta\_dot')

% Plotting the acceleration of theta
figure
plot(theta_interval(2:end),D2theta)
title('theta\_dot\_dot')

% Plotting the reduced dynamics equation (should be close to zero)
figure
plot(theta_interval(2:end),red_eq)
title('Approximation of reduced system dynamics')
xlabel('theta')

%% This section MUST be run before running any code below!
% This section is for solving the 2nd-order differential equation of the
% reduced dynamics.

VHC_coeff = z0;

options = odeset('RelTol',1e-9,'AbsTol',1e-9,'MaxStep',1e-3);
time_span = [0 0.6];

init = [theta_b 0];
[t_d,x] = ode45(@reduced_dynamics,time_span,init,options);
theta_d = x(:,1);
Dtheta_d = x(:,2);

% Finding the relevant interval of phase-portrait
for i = 1:length(theta_d)
    if abs(theta_d(i)-theta_e)<0.001
        theta_d = theta_d(1:i);
        Dtheta_d = Dtheta_d(1:i);
        t_d = t_d(1:i);
        break
    end
end

plot(theta_d,Dtheta_d)
hold on
title('Phase-plot')
ylabel('$\dot{\theta}$','interpreter','latex')
xlabel('\theta')

% With respect to time:
% plot(t_d,theta_d)  
% plot(t_d,Dtheta_d)

%% JOINT ANGLES
[q1_d,q2_d] = phi(theta_d,VHC_coeff);
q3_d = acos((theta_d-l1*cos(q1_d-psi)-l2*cos(q1_d+q2_d-psi))./l3)-q1_d-q2_d+psi;

figure
subplot(3,1,1)
plot(t_d,q1_d)
ylabel('q_1(t) [rad]')
title('Joint Angles')

subplot(3,1,2)
plot(t_d,q2_d)
ylabel('q_2(t) [rad]')

subplot(3,1,3)
plot(t_d,q3_d)
ylabel('q_3(t) [rad]')
xlabel('time[s]')

%% JOINT VELOCITIES
[DPhi1,DPhi2] = Dphi(theta_d,VHC_coeff);
Dq1_d = DPhi1.*Dtheta_d;
Dq2_d = DPhi2.*Dtheta_d;
Dq3_d = (-Dtheta_d-l1.*sin(q1_d-psi).*Dq1_d-l2.*sin(q1_d+q2_d-psi).*(Dq1_d+Dq2_d))./(l3.*sin(q1_d+q2_d+q3_d-psi))-Dq1_d-Dq2_d;

figure
subplot(3,1,1)
plot(t_d,Dq1_d)
ylabel('$\dot{q}_1(t)$ \ [rad/s]','interpreter','latex')
line([0,t_d(end)],[-11/18*pi,-11/18*pi],'Linewidth',2,'Color',[0,0,0])
line([0,t_d(end)],[11/18*pi,11/18*pi],'Linewidth',2,'Color',[0,0,0])
ylim([-2.2,2.2])
title('Joint Velocities')

subplot(3,1,2)
plot(t_d,Dq2_d)
ylabel('$\dot{q}_2(t)$  \ [rad/s]','interpreter','latex')
line([0,t_d(end)],[-32/45*pi,-32/45*pi],'Linewidth',2,'Color',[0,0,0])
line([0,t_d(end)],[32/45*pi,32/45*pi],'Linewidth',2,'Color',[0,0,0])
ylim([-2.5,2.5])

subplot(3,1,3)
plot(t_d,Dq3_d)
ylabel('$\dot{q}_3(t)$ \ [rad/s]','interpreter','latex')
xlabel('time[s]')
line([0,t_d(end)],[-46/45*pi,-46/45*pi],'Linewidth',2,'Color',[0,0,0])
line([0,t_d(end)],[46/45*pi,46/45*pi],'Linewidth',2,'Color',[0,0,0])

%% TORQUES
[Torque1_d,Torque2_d] = torque(theta_d,VHC_coeff,Dtheta_d.^2);

figure
subplot(2,1,1)
plot(t_d,Torque1_d)
ylabel('$\tau_1(t) \ [Nm]$','interpreter','latex')
line([0,t_d(end)],[-176,-176],'Linewidth',2,'Color',[0,0,0])
line([0,t_d(end)],[176,176],'Linewidth',2,'Color',[0,0,0])
ylim([-200,200])
title('Torque')

subplot(2,1,2)
plot(t_d,Torque2_d)
ylabel('$\tau_2(t) \ [Nm]$','interpreter','latex')
xlabel('time[s]')
line([0,t_d(end)],[-100,-100],'Linewidth',2,'Color',[0,0,0])
line([0,t_d(end)],[100,100],'Linewidth',2,'Color',[0,0,0])
ylim([-120,120])

%% DISTANCE THROWN
x_e = l1*cos(q1_d(end))+l2*cos(q1_d(end)+q2_d(end))+l3*cos(q1_d(end)+q2_d(end)+q3_d(end));
y_e = l0+l1*sin(q1_d(end))+l2*sin(q1_d(end)+q2_d(end))+l3*sin(q1_d(end)+q2_d(end)+q3_d(end))+0.63; 
% 0.63 is the height of the platform the robot is placed on.

x_vel = Dtheta_d(end)*cos(psi);
y_vel = Dtheta_d(end)*sin(psi);

d = x_vel*y_vel/g+sqrt((x_vel*y_vel/g)^2+2*x_vel^2*y_e/g)+x_e

%% Plot for position control only

% Positions
figure
subplot(3,1,1)
plot(t_d,q1_d,Time_pos,q1_pos)
legend('model','robot')
ylabel('[rad]')
title('q_1')
subplot(3,1,2) 
plot(t_d,q2_d,Time_pos,q2_pos)
ylabel('[rad]')
title('q_2')
subplot(3,1,3)
plot(t_d,q3_d,Time_pos,q3_pos)
ylabel('[rad]')
xlabel('[s]')
title('q_3')

% Velocity
q1_vel = diff(q1_pos)./diff(Time_pos);
q2_vel = diff(q2_pos)./diff(Time_pos);
q3_vel = diff(q3_pos)./diff(Time_pos);
figure
subplot(3,1,1)
plot(t_d,Dq1_d,Time_pos(2:end),q1_vel)
line([0,t_d(end)],[-11/18*pi,-11/18*pi],'Linewidth',2,'Color',[0,0,0])
legend('model','robot')
ylabel('[rad/s]')
title('$\dot{q}_1$ ','interpreter','latex')
subplot(3,1,2)
plot(t_d,Dq2_d,Time_pos(2:end),q2_vel)
line([0,t_d(end)],[-32/45*pi,-32/45*pi],'Linewidth',2,'Color',[0,0,0])
ylabel('[rad/s]')
title('$\dot{q}_2$ ','interpreter','latex')
subplot(3,1,3)
plot(t_d,Dq3_d,Time_pos(2:end),q3_vel)
line([0,t_d(end)],[-46/45*pi,-46/45*pi],'Linewidth',2,'Color',[0,0,0])
ylabel('[rad/s]')
xlabel('[s]')
title('$\dot{q}_3$ ','interpreter','latex')

% Torque
figure
subplot(2,1,1)
plot(t_d,Torque1_d,Time_pos,Torque1_pos)
legend('model','robot')
ylabel('[Nm]')
title('$\tau_1$ ','interpreter','latex')
subplot(2,1,2)
plot(t_d,Torque2_d,Time_pos,Torque2_pos)
ylabel('[Nm]')
xlabel('[s]')
title('$\tau_2$ ','interpreter','latex')

% Desired positions
figure
plot(t_d,q1_d,Time_pos,Desired_q1()
legend('model','robot')
ylabel('[rad]')
xlabel('[s]')
title('q_1')
figure
plot(t_d,q2_d,Time_pos,Desired_q2)
legend('model','robot')
ylabel('[rad]')
xlabel('[s]')
title('q_2')
figure
plot(t_d,q3_d,Time_pos,Desired_q3)
legend('model','robot')
ylabel('[rad]')
xlabel('[s]')
title('q_3')

%% Plot for Position and Torque control

% Positions
figure
subplot(3,1,1)
plot(t_d,q1_d,Time,q1_trq)
legend('model','robot')
ylabel('[rad]')
title('q_1')
subplot(3,1,2)
plot(t_d,q2_d,Time,q2_trq)
ylabel('[rad]')
title('q_2')
subplot(3,1,3)
plot(t_d,q3_d,Time,q3_trq)
ylabel('[rad]')
xlabel('[s]')
title('q_3')

% Velocity
q1_vel = diff(q1_trq)./diff(Time);
q2_vel = diff(q2_trq)./diff(Time);
q3_vel = diff(q3_trq)./diff(Time);
figure
subplot(3,1,1)
plot(t_d,Dq1_d,Time(2:end),q1_vel)
line([0,t_d(end)],[-11/18*pi,-11/18*pi],'Linewidth',2,'Color',[0,0,0])
legend('model','robot')
ylabel('[rad/s]')
title('$\dot{q}_1$ ','interpreter','latex')
subplot(3,1,2)
plot(t_d,Dq2_d,Time(2:end),q2_vel)
line([0,t_d(end)],[-32/45*pi,-32/45*pi],'Linewidth',2,'Color',[0,0,0])
ylabel('[rad/s]')
title('$\dot{q}_2$ ','interpreter','latex')
subplot(3,1,3)
plot(t_d,Dq3_d,Time(2:end),q3_vel)
line([0,t_d(end)],[-46/45*pi,-46/45*pi],'Linewidth',2,'Color',[0,0,0])
ylabel('[rad/s]')
xlabel('[s]')
title('$\dot{q}_3$ ','interpreter','latex')

% Torque
figure
subplot(2,1,1)
plot(t_d,Torque1_d,Time,Torque1_trq)
legend('model','robot')
ylabel('[Nm]')
title('$\tau_1$ ','interpreter','latex')
subplot(2,1,2)
plot(t_d,Torque2_d,Time,Torque2_trq)
ylabel('[Nm]')
xlabel('[s]')
title('$\tau_2$ ','interpreter','latex')

% Desired positions
figure
plot(t_d,q1_d,Time,Desired_q1)
legend('model','robot')
ylabel('[rad]')
xlabel('[s]')
title('q_1')
figure
plot(t_d,q2_d,Time,Desired_q2)
legend('model','robot')
ylabel('[rad]')
xlabel('[s]')
title('q_2')
figure
plot(t_d,q3_d,Time,Desired_q3)
legend('model','robot')
ylabel('[rad]')
xlabel('[s]')
title('q_3')
%% Saving data to text-file as input data for KUKA lwr
% Saving with a sample-time of 2ms

j = 8;
h = 1;
for i = 1:length(t_d)
    if j == 8
       
        time(h) = t_d(i);
        q1(h) = q1_d(i);
        q2(h) = q2_d(i);
        q3(h) = q3_d(i);
        
        torque1(h) = Torque1_d(i);
        torque2(h) = Torque2_d(i);
        torque3(h) = 0;
        
        j = 0;
        h = h+1;
    end
    j = j+1;
end


fileID = fopen('pitch2.txt','w');
fprintf(fileID,'%4.4f %4.4f %4.4f %4.4f %4.4f %4.4f\r\n',[q1; q2; q3; torque1; torque2; torque3]);